==> magento-module-saas-order-sync-1.2.0.zip: Core/OrderSync/Manager.php <==
<?php
/************************************************************************
 *
 * ADOBE CONFIDENTIAL
 * ___________________
 *
 * Copyright 2023 Adobe
 * All Rights Reserved.
 *
 * NOTICE: All information contained herein is, and remains
 * the property of Adobe and its suppliers, if any. The intellectual
 * and technical concepts contained herein are proprietary to Adobe
 * and its suppliers and are protected by all applicable intellectual
 * property laws, including trade secret and copyright laws.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Adobe.
 * ************************************************************************
 */
declare(strict_types=1);

namespace Magento\SaaSOrderSync\Core\OrderSync;

use Symfony\Component\Stopwatch\Stopwatch;
use function __;
use DateTime;
use DateTimeInterface;
use Exception;
use Magento\Framework\DataObject\IdentityGeneratorInterface;
use Magento\Framework\Webapi\Exception as WebapiException;
use Magento\SaaSOrderSync\Api\OrderSync\CancelOrderSyncResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\CreateOrderSyncResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\GetAllOrderSyncsResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\GetOneOrderSyncResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\OrderSyncManagerInterface;
use Magento\SaaSOrderSync\Core\CommerceDataExport\OrdersDataExporterAdapter;
use Magento\SaaSOrderSync\Core\Common\Time\ClockInterface;
use Magento\SaaSOrderSync\Core\OrderSync\Bulk\Manager as BulkManager;
use Magento\SaaSOrderSync\Core\SaaS\SaaSClientResolverInterface;
use Psr\Log\LoggerInterface;
use Throwable;

class Manager implements OrderSyncManagerInterface
{
    private IdentityGeneratorInterface $identityGenerator;
    private OrdersDataExporterAdapter $ordersExporter;
    private BulkManager $ordersBulkManager;
    private SaaSClientResolverInterface $clientResolver;
    private LoggerInterface $logger;
    private Lock $orderSyncLock;
    private ClockInterface $clock;

    public function __construct(
        IdentityGeneratorInterface  $identityGenerator,
        OrdersDataExporterAdapter   $ordersExporter,
        BulkManager                 $ordersBulkManager,
        SaaSClientResolverInterface $clientResolver,
        LoggerInterface             $logger,
        Lock                        $orderSyncLock,
        ClockInterface              $clock,
    ) {
        $this->identityGenerator = $identityGenerator;
        $this->ordersExporter = $ordersExporter;
        $this->ordersBulkManager = $ordersBulkManager;
        $this->clientResolver = $clientResolver;
        $this->logger = $logger;
        $this->orderSyncLock = $orderSyncLock;
        $this->clock = $clock;
    }

    public function createOrderSync(
        string $created_from,
        string $client_code,
        bool   $optimize_time_range = true,
        bool   $force = false,
    ): CreateOrderSyncResponseInterface {
        $syncId = $this->identityGenerator->generateId();
        $ctx = ['syncId' => $syncId, 'createdFrom' => $created_from, 'clientCode' => $client_code];

        $createdFrom = $this->validateCreatedFrom($created_from);
        $clientCode = $this->parseString($client_code)
            ?: $this->throwBadRequest('client_code can\'t be empty.');

        if (!$this->ordersBulkManager->isAMQPAvailable()) {
            throw new WebapiException(
                __(
                    'Can\'t start Order Sync due to missing AMQP connection. Request to the administrator of the'
                    . ' Adobe Commerce instance to configure Message Queue Framework accordingly.'
                    . ' See: https://experienceleague.adobe.com/docs/commerce-operations/configuration-guide/message-queues/message-queue-framework.html?lang=en'
                ),
                httpCode: WebapiException::HTTP_INTERNAL_ERROR,
            );
        }

        if (!$optimize_time_range) {
            $this->logger->warning('Create order sync with optimize_time_range=false is deprecated, use force=true instead.');
        }

        try {
            $now = $this->clock->now();
            $syncTimeFrame = $this->resolveSyncTimeRange($createdFrom, $now, !$force);
            if (empty($syncTimeFrame)) {
                return $this->createCompletedSync($syncId, $createdFrom, $now, $clientCode);
            }

            [$createdFrom, $createdTo] = $syncTimeFrame;

            $initialCount = $this->ordersExporter->count($createdFrom, $createdTo);
            if ($initialCount == 0) {
                return $this->createCompletedSync($syncId, $createdFrom, $createdTo, $clientCode);
            }

            $orderSyncClient = $this->clientResolver->createOrderSyncClient();
            $orderSync = $orderSyncClient->createOrderSync(
                $syncId,
                $createdFrom,
                $createdTo,
                $initialCount,
                $clientCode,
                $force
            );

            $stopwatch = new Stopwatch(true);
            $stopwatch->start($syncId);
            $this->ordersBulkManager->schedule($syncId, $createdTo, $createdFrom);
            $creationMillis = $stopwatch->stop($syncId)->getDuration();

            $this->logger->info("Created SaaS order sync, took ${creationMillis}ms.", $ctx);

            return new CreateOrderSyncResponse($this->map($orderSync));
        } catch (WebapiException $e) {
            throw $e;
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".", $ctx);
        }
    }

    public function getAllOrderSyncs(): GetAllOrderSyncsResponseInterface
    {
        try {
            $orderSyncClient = $this->clientResolver->createOrderSyncClient();
            ['orderSyncs' => $orderSyncs] = $orderSyncClient->getAllOrderSyncs();

            // @formatter:off
            return new GetAllOrderSyncsResponse(array_map(
                fn (array $orderSync): OrderSync => $this->map($orderSync),
                $orderSyncs
            ));
            // @formatter:on
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".");
        }
    }

    public function getOneOrderSync($sync_id): GetOneOrderSyncResponseInterface
    {
        try {
            $orderSyncClient = $this->clientResolver->createOrderSyncClient();
            $orderSync = $orderSyncClient->getOneOrderSync($sync_id);

            if ($orderSync == null) {
                throw new WebapiException(__("No order sync found for id '$sync_id'."));
            }

            return new GetOneOrderSyncResponse($this->map($orderSync));
        } catch (WebapiException $e) {
            throw $e;
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".", ['syncId' => $sync_id]);
        }
    }

    public function cancelOrderSync(string $sync_id): CancelOrderSyncResponseInterface
    {
        try {
            return $this->orderSyncLock->locking($sync_id, function ($syncId) {
                $orderSyncClient = $this->clientResolver->createOrderSyncClient();
                [$err, $data] = $orderSyncClient->cancelOrderSync($syncId);

                if ($err) {
                    match ($err['code']) {
                        'NOT_FOUND' => throw new WebapiException(__("No order sync found for id '$syncId'.")),
                        'NOT_IN_PROGRESS' => throw new WebapiException(__("Order sync '$syncId' is not in progress."), httpCode: 409),
                        default => $this->logAndThrow(
                            'Unexpected error while canceling order sync in SaaS.',
                            ['syncId' => $syncId, 'error' => $err]
                        )
                    };
                }

                $this->ordersBulkManager->cancel($syncId);

                $this->logger->info('SaaS order sync has been canceled.', ['syncId' => $syncId]);

                return new CancelOrderSyncResponse($this->map($data));
            });
        } catch (WebapiException $e) {
            throw $e;
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".", ['syncId' => $sync_id]);
        }
    }

    private function map(array $orderSync): OrderSync
    {
        return new OrderSync(
            $orderSync['syncId'],
            $orderSync['createdFrom'],
            $orderSync['createdTo'],
            $orderSync['clientCode'],
            $orderSync['initialCount'],
            $orderSync['processedCount'],
            $orderSync['errorCount'],
            $orderSync['status']
        );
    }

    private function validateCreatedFrom(string $createdFrom): DateTime
    {
        $parsedCreatedFrom = $this->parseDateTime($createdFrom)
            ?: $this->throwBadRequest('created_from can\'t be parsed in ISO8601 format.');

        if ($this->clock->now() < $parsedCreatedFrom) {
            $this->throwBadRequest('created_from should be in the past.');
        }

        return $parsedCreatedFrom;
    }

    public function parseDateTime(string $dateTime): DateTime|bool
    {
        try {
            return new DateTime($dateTime);
        } catch (Exception) {
            return false;
        }
    }

    private function parseString(string $string): string|bool
    {
        return strlen(trim($string)) === 0 ? false : $string;
    }

    private function resolveSyncTimeRange(DateTime $createdFrom, DateTime $now, bool $optimize_time_range): array
    {
        if (!$optimize_time_range) {
            return [$createdFrom, $now];
        }

        $merchantClient = $this->clientResolver->createMerchantClient();
        ['orders' => $ordersDataSummary] = $merchantClient->getDataSummary();
        if (!array_key_exists('createdFrom', $ordersDataSummary)) {
            return [$createdFrom, $now];
        }

        $createdTo = $this->parseDateTime($ordersDataSummary['createdFrom']);
        if (!$createdTo) {
            $this->logAndThrow(
                'orders created_from of merchant data summary can\'t be parsed in ISO8601 format.',
                ['ordersDataSummary' => $ordersDataSummary]
            );
        }

        if ($createdFrom >= $createdTo) {
            $this->logger->debug('Order sync time range'
                . " from {$createdFrom->format(DateTimeInterface::ATOM)}"
                . " to {$createdTo->format(DateTimeInterface::ATOM)}"
                . ' is already synced.');
            return [];
        }

        return [$createdFrom, $createdTo];
    }

    private function createCompletedSync(string $syncId, DateTime $createdFrom, DateTime $createdTo, string $clientCode): CreateOrderSyncResponseInterface
    {
        $orderSyncClient = $this->clientResolver->createOrderSyncClient();
        $orderSyncClient->createOrderSync($syncId, $createdFrom, $createdTo, 0, $clientCode);
        [$err, $data] = $orderSyncClient->completeOrderSync($syncId);
        if ($err) {
            $this->logAndThrow('Failed to complete order sync.', ['syncId' => $syncId]);
        }

        return new CreateOrderSyncResponse($this->map($data));
    }

    private function logAndThrow(string $msg, array $ctx = [])
    {
        $e = new WebapiException(
            __($msg),
            httpCode: WebapiException::HTTP_INTERNAL_ERROR
        );
        $this->logAndRethrow($e, $msg, $ctx);
    }

    private function logAndRethrow(Throwable $e, string $msg, array $ctx = [])
    {
        $ctx['exception'] = $e;
        $this->logger->critical($msg, $ctx);
        throw $e;
    }

    private function throwBadRequest($msg)
    {
        throw new WebapiException(
            __($msg),
            httpCode: WebapiException::HTTP_BAD_REQUEST,
        );
    }
}

==> magento-module-saas-order-sync-1.4.0.zip: Core/OrderSync/Manager.php <==
<?php
/************************************************************************
 *
 * ADOBE CONFIDENTIAL
 * ___________________
 *
 * Copyright 2023 Adobe
 * All Rights Reserved.
 *
 * NOTICE: All information contained herein is, and remains
 * the property of Adobe and its suppliers, if any. The intellectual
 * and technical concepts contained herein are proprietary to Adobe
 * and its suppliers and are protected by all applicable intellectual
 * property laws, including trade secret and copyright laws.
 * Dissemination of this information or reproduction of this material
 * is strictly forbidden unless prior written permission is obtained
 * from Adobe.
 * ************************************************************************
 */
declare(strict_types=1);

namespace Magento\SaaSOrderSync\Core\OrderSync;

use DateTime;
use Exception;
use Magento\Framework\DataObject\IdentityGeneratorInterface;
use Magento\Framework\Webapi\Exception as WebapiException;
use Magento\SaaSOrderSync\Api\OrderSync\CancelOrderSyncResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\CreateOrderSyncResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\GetAllOrderSyncsResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\GetOneOrderSyncResponseInterface;
use Magento\SaaSOrderSync\Api\OrderSync\OrderSyncManagerInterface;
use Magento\SaaSOrderSync\Core\CommerceDataExport\OrdersDataExporterAdapter;
use Magento\SaaSOrderSync\Core\Common\Time\ClockInterface;
use Magento\SaaSOrderSync\Core\OrderSync\Bulk\Manager as BulkManager;
use Magento\SaaSOrderSync\Core\SaaS\SaaSClientResolverInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\Stopwatch\Stopwatch;
use Throwable;
use function __;

class Manager implements OrderSyncManagerInterface
{
    private IdentityGeneratorInterface $identityGenerator;
    private OrdersDataExporterAdapter $ordersExporter;
    private BulkManager $ordersBulkManager;
    private SaaSClientResolverInterface $clientResolver;
    private LoggerInterface $logger;
    private Lock $orderSyncLock;
    private ClockInterface $clock;

    public function __construct(
        IdentityGeneratorInterface  $identityGenerator,
        OrdersDataExporterAdapter   $ordersExporter,
        BulkManager                 $ordersBulkManager,
        SaaSClientResolverInterface $clientResolver,
        LoggerInterface             $logger,
        Lock                        $orderSyncLock,
        ClockInterface              $clock,
    ) {
        $this->identityGenerator = $identityGenerator;
        $this->ordersExporter = $ordersExporter;
        $this->ordersBulkManager = $ordersBulkManager;
        $this->clientResolver = $clientResolver;
        $this->logger = $logger;
        $this->orderSyncLock = $orderSyncLock;
        $this->clock = $clock;
    }

    public function createOrderSync(
        string $client_code,
        string $created_from,
        string $created_to = 'now'
    ): CreateOrderSyncResponseInterface {
        $syncId = $this->identityGenerator->generateId();
        $ctx = [
            'syncId' => $syncId,
            'clientCode' => $client_code,
            'createdFrom' => $created_from,
            'createdTo' => $created_to
        ];

        $clientCode = $this->parseString($client_code)
            ?: $this->throwBadRequest('client_code can\'t be empty.');
        $createdFrom = $this->validateCreatedFrom($created_from);
        $createdTo = $this->validateCreatedTo($created_to);

        if (!$this->ordersBulkManager->isAMQPAvailable()) {
            throw new WebapiException(
                __(
                    'Can\'t start Order Sync due to missing AMQP connection. Request to the administrator of the'
                    . ' Adobe Commerce instance to configure Message Queue Framework accordingly.'
                    . ' See: https://experienceleague.adobe.com/docs/commerce-operations/configuration-guide/message-queues/message-queue-framework.html?lang=en'
                ),
                httpCode: WebapiException::HTTP_INTERNAL_ERROR,
            );
        }

        try {
            $initialCount = $this->ordersExporter->count($createdFrom, $createdTo);
            if ($initialCount == 0) {
                return $this->createCompletedSync($syncId, $clientCode, $createdFrom, $createdTo);
            }

            $orderSyncClient = $this->clientResolver->createOrderSyncClient();
            $orderSync = $orderSyncClient->createOrderSync(
                $syncId,
                $createdFrom,
                $createdTo,
                $initialCount,
                $clientCode
            );

            $stopwatch = new Stopwatch(true);
            $stopwatch->start($syncId);
            $this->ordersBulkManager->schedule($syncId, $createdTo, $createdFrom);
            $creationMillis = $stopwatch->stop($syncId)->getDuration();

            $this->logger->info("Created SaaS order sync, took {$creationMillis}ms.", $ctx);

            return new CreateOrderSyncResponse($this->map($orderSync));
        } catch (WebapiException $e) {
            throw $e;
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".", $ctx);
        }
    }

    public function getAllOrderSyncs(): GetAllOrderSyncsResponseInterface
    {
        try {
            $orderSyncClient = $this->clientResolver->createOrderSyncClient();
            ['orderSyncs' => $orderSyncs] = $orderSyncClient->getAllOrderSyncs();

            // @formatter:off
            return new GetAllOrderSyncsResponse(array_map(
                fn (array $orderSync): OrderSync => $this->map($orderSync),
                $orderSyncs
            ));
            // @formatter:on
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".");
        }
    }

    public function getOneOrderSync($sync_id): GetOneOrderSyncResponseInterface
    {
        try {
            $orderSyncClient = $this->clientResolver->createOrderSyncClient();
            $orderSync = $orderSyncClient->getOneOrderSync($sync_id);

            if ($orderSync == null) {
                throw new WebapiException(__("No order sync found for id '$sync_id'."));
            }

            return new GetOneOrderSyncResponse($this->map($orderSync));
        } catch (WebapiException $e) {
            throw $e;
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".", ['syncId' => $sync_id]);
        }
    }

    public function cancelOrderSync(string $sync_id): CancelOrderSyncResponseInterface
    {
        try {
            return $this->orderSyncLock->locking($sync_id, function ($syncId) {
                $orderSyncClient = $this->clientResolver->createOrderSyncClient();
                [$err, $data] = $orderSyncClient->cancelOrderSync($syncId);

                if ($err) {
                    match ($err['code']) {
                        'NOT_FOUND' => throw new WebapiException(__("No order sync found for id '$syncId'.")),
                        'NOT_IN_PROGRESS' => throw new WebapiException(__("Order sync '$syncId' is not in progress."), httpCode: 409),
                        default => $this->logAndThrow(
                            'Unexpected error while canceling order sync in SaaS.',
                            ['syncId' => $syncId, 'error' => $err]
                        )
                    };
                }

                $this->ordersBulkManager->cancel($syncId);

                $this->logger->info('SaaS order sync has been canceled.', ['syncId' => $syncId]);

                return new CancelOrderSyncResponse($this->map($data));
            });
        } catch (WebapiException $e) {
            throw $e;
        } catch (Throwable $e) {
            $this->logAndRethrow($e, "Unexpected exception on " . (__METHOD__) . ".", ['syncId' => $sync_id]);
        }
    }

    private function map(array $orderSync): OrderSync
    {
        return new OrderSync(
            $orderSync['syncId'],
            $orderSync['createdFrom'],
            $orderSync['createdTo'],
            $orderSync['clientCode'],
            $orderSync['initialCount'],
            $orderSync['processedCount'],
            $orderSync['errorCount'],
            $orderSync['status']
        );
    }

    private function validateCreatedFrom(string $createdFrom): DateTime
    {
        $parsedCreatedFrom = $this->parseDateTime($createdFrom)
            ?: $this->throwBadRequest('created_from can\'t be parsed in ISO8601 format.');

        if ($this->clock->now() < $parsedCreatedFrom) {
            $this->throwBadRequest('created_from should be in the past.');
        }

        return $parsedCreatedFrom;
    }

    private function validateCreatedTo(string $createdTo): DateTime
    {
        if ($createdTo === 'now') {
            return new DateTime('now');
        }

        return $this->parseDateTime($createdTo)
            ?: $this->throwBadRequest('created_to can\'t be parsed in ISO8601 format.');
    }

    public function parseDateTime(string $dateTime): DateTime|bool
    {
        try {
            return new DateTime($dateTime);
        } catch (Exception) {
            return false;
        }
    }

    private function parseString(string $string): string|bool
    {
        return strlen(trim($string)) === 0 ? false : $string;
    }

    private function createCompletedSync(
        string   $syncId,
        string   $clientCode,
        DateTime $createdFrom,
        DateTime $createdTo
        ): CreateOrderSyncResponseInterface
    {
        $orderSyncClient = $this->clientResolver->createOrderSyncClient();
        $orderSyncClient->createOrderSync($syncId, $createdFrom, $createdTo, 0, $clientCode);
        [$err, $data] = $orderSyncClient->completeOrderSync($syncId);
        if ($err) {
            $this->logAndThrow('Failed to complete order sync.', ['syncId' => $syncId]);
        }

        return new CreateOrderSyncResponse($this->map($data));
    }

    private function logAndThrow(string $msg, array $ctx = [])
    {
        $e = new WebapiException(
            __($msg),
            httpCode: WebapiException::HTTP_INTERNAL_ERROR
        );
        $this->logAndRethrow($e, $msg, $ctx);
    }

    private function logAndRethrow(Throwable $e, string $msg, array $ctx = [])
    {
        $ctx['exception'] = $e;
        $this->logger->critical($msg, $ctx);
        throw $e;
    }

    private function throwBadRequest($msg)
    {
        throw new WebapiException(
            __($msg),
            httpCode: WebapiException::HTTP_BAD_REQUEST,
        );
    }
}
